@@ -5,6 +5,7 @@ import ast
 import collections
 import inspect
 import itertools
+import linecache
 import logging
 import operator
 import os.path
@@ -14,6 +15,9 @@ import sys
 import typing
 import warnings
 
+# Ignore, at least, these directory basenames (default directories to ignore).
+default_ignore_dirs = (".venv", ".env", ".git", ".hg", ".pytest_cache", ".ruff_cache", "__pycache__")
+
 # Docstring sections: https://www.sphinx-doc.org/en/master/usage/domains/python.html#info-field-lists
 ptype_key = "type"
 rtype_key = "rtype"
@@ -32,6 +36,8 @@ module_var_key = "var"
 vartype_key = "vartype"
 ignore_set = {"meta"}  # At this moment, it's ignored
 
+types_set = {ptype_key, rtype_key, vartype_key}  # Sections that define types
+
 # Summary (everything before the first section or the end of the string)
 summary_regex = re.compile(r'^(.*?)(?=^:|\Z)', flags=re.DOTALL | re.MULTILINE)
 # Section (start with ':' and end before next ':' at start of line or end of string)
@@ -43,6 +49,8 @@ empty_lines_regex = re.compile("(?:^[ \t]*\r?\n){2,}(?=[^\r\n])", re.MULTILINE)
 # Docstring starting or ending with only quotes lines
 quotes_starts_regex = re.compile(r'^"+\s*$')
 quotes_ends_regex = re.compile(r'^\s*"+$')
+# Bad whitespace in section definition: (consecutive spaces or trailing space or leading space)
+section_bad_ws_regex = re.compile(r'(\s{2,}|^\s+|\s+$)')
 
 
 class NodeTypes:
@@ -94,12 +102,15 @@ class ParsedDocs(typing.NamedTuple):
     variables: list[ParsedDocsVar]
     invalid: list[str]  # invalid section keys
     ignored: list[str]  # ignored section keys
+    bad_whitespaces_def: list[str]  # section definitions with bad whitespaces
 
     rawdocs: str | None  # Raw docstring
     docs: str | None  # Cleaned docstring (inspect.cleandoc)
     docs_ini_lineno: int | None  # First line number of the docstring, None if no docstring
     docs_end_lineno: int | None  # End line number of the docstring, None if no docstring
-    code_ini_lineno: int | None  # First line number of the code block after the docstring, None if no code
+    code_ini_lineno: int | None  # First line number of the code or comment after the docstring.
+
+    non_empty_end_lines: bool  # True if there are non-empty lines after the last section
 
 
 class Violations:
@@ -111,9 +122,9 @@ class Violations:
     DOC002 = (True, "DOC002", "Malformed section ({!r})")
     DOC003 = (True, "DOC003", "Missing blank line after docstring")
     DOC004 = (True, "DOC004", "Missing blank line between summary and sections")
-    DOC005 = (True, "DOC005", "Too many consecutive empty lines")
+    DOC005 = (True, "DOC005", "Too many consecutive blank lines")
     DOC006 = (True, "DOC006", "Trailing empty lines")
-    DOC007 = (True, "DOC007", "Misplaced section ({!r} after {!r})")
+    DOC007 = (True, "DOC007", "Misplaced section ({!r} appears after {!r})")
     # Ruff (missing-trailing-period) → enforces a trailing period on all docstring summaries (one-line and multi-line).
     # Rule (DOC008) → only enforces a trailing period on one-line docstrings, as recommended by PEP257.
     DOC008 = (True, "DOC008", "One-line docstring should end with a period")
@@ -121,15 +132,18 @@ class Violations:
     # this rule only checks that multi-line docstrings do not start or end with more than three double quotes.
     DOC009 = (True, "DOC009", "Docstring must not use more than 3 double quotes")
 
+    DOC010 = (True, "DOC010", "Section definition contains invalid whitespace ({!r})")
+    DOC011 = (True, "DOC011", "Trailing non-empty lines after last section.")
+
     # DOC1xx: Parameter issues
     DOC101 = (True, "DOC101", "Parameter documented but not in signature ({!r})")
     DOC102 = (True, "DOC102", "Invalid parameter type syntax ({!r})")
     DOC103 = (True, "DOC103", "Parameter type already in signature ({!r})")
-    DOC104 = (True, "DOC104", "Parameter type mismatch with hint ({!r} != {!r})")
+    DOC104 = (True, "DOC104", "Parameter type mismatch with annotation ({!r} != {!r})")
     DOC105 = (True, "DOC105", "Duplicated parameter ({!r})")
 
     # DOC2xx: Return issues
-    DOC201 = (True, "DOC201", "Return documented but function has no return")
+    DOC201 = (True, "DOC201", "Return documented but function has no return statement")
     DOC202 = (True, "DOC202", "Invalid return type syntax ({!r})")
     DOC203 = (True, "DOC203", "Return type already in signature ({!r})")
     DOC204 = (True, "DOC204", "Return type mismatch with annotation ({!r} != {!r})")
@@ -141,7 +155,7 @@ class Violations:
 
     # DOC4xx: Variable issues
     DOC402 = (True, "DOC402", "Invalid variable type syntax ({!r})")
-    DOC403 = (True, "DOC403", "Variable name must not contain spaces ({!r})")
+    DOC403 = (True, "DOC403", "Variable name contains invalid whitespace ({!r})")
     DOC405 = (True, "DOC405", "Duplicated variable ({!r})")
 
     _get_order = operator.attrgetter("order")
@@ -192,6 +206,11 @@ class Violations:
                     # finditer: Empty matches are included in the result.
                     yield cls.DOC006, ()
 
+    @classmethod
+    def validate_bad_whitespaces(cls, parsed, /):
+        for section_def in parsed.bad_whitespaces_def:
+            yield cls.DOC010, (section_def,)
+
     @classmethod
     def validate_head_tail_quotes(cls, parsed, /):
         text = parsed.rawdocs
@@ -341,6 +360,10 @@ class Violations:
         yield from cls.validate_empty_lines(parsed)
         yield from cls.validate_head_tail_quotes(parsed)
         yield from cls.validate_summary(parsed)
+        yield from cls.validate_bad_whitespaces(parsed)
+
+        if parsed.non_empty_end_lines:
+            yield cls.DOC011, ()
 
         if parsed.kind == NodeTypes.FUNCTION:  # Only functions have parameters, returns, and raises
             yield from cls.validate_params(parsed, parameters)
@@ -460,7 +483,33 @@ def get_node_type(node, /):
     return kind
 
 
-def parse_docs(node, /):
+def fetch_bad_ws_definitions(section, section_key, sep, a_raw, b_raw, /):
+    """
+    Yields the parts of the section definition that have bad whitespaces.
+
+    :param str section: Full section text.
+    :param str section_key: Section key (first word of part before ':').
+    :param bool sep: True if ':' was present.
+    :param str a_raw: Part before the first ':'
+    :param str b_raw: Part after the first ':'
+
+    :return: Generator yielding section definitions with bad whitespaces.
+    """
+
+    if section_key in types_set:  # Check both sides for type sections
+        # Get right side (after '') without leading spaces and split into lines
+        right_side_lines = b_raw.removeprefix(" ").splitlines()
+        # Only first line of right side is relevant.
+        right_side = right_side_lines[0] if right_side_lines else ""
+        if section_bad_ws_regex.search(a_raw) or section_bad_ws_regex.search(right_side):
+            yield section.splitlines()[0]  # Return only the section definition line
+
+    elif section_bad_ws_regex.search(a_raw):  # Other sections only check left-side
+        # Reconstruct section definition with only left-side
+        yield ":{}{}".format(a_raw, ":" if sep else "")
+
+
+def parse_docs(node, filename, /):
     params = list()
     raises = list()
     returns = list()
@@ -469,6 +518,7 @@ def parse_docs(node, /):
     # Does not use 'set' to preserve the order of appearance
     invalid = list()
     ignored = list()
+    bad_whitespaces_def = list()  # List of section definitions with bad whitespaces
 
     rawdocs = ast.get_docstring(node, clean=False)
     docs = rawdocs if rawdocs is None else inspect.cleandoc(rawdocs)
@@ -477,15 +527,23 @@ def parse_docs(node, /):
     is_func = kind == NodeTypes.FUNCTION
     is_class = kind == NodeTypes.CLASS
     is_module = kind == NodeTypes.MODULE
+
+    section_key, b_raw = (None, None)  # Initialize for misplaced_tail_text check
     for order, section in enumerate(itersections(docs)):
-        a, sep, b = section.lstrip(":").partition(":")
+        a_raw, sep, b_raw = section.lstrip(":").partition(":")
         sep = bool(sep)  # True if ':' was present
-        b = b.splitlines()[0].strip() if (b := b.strip()) else ""  # Only first line of description is relevant.
-        parts_a = a.split()
+        b = b.splitlines()[0].strip() if (b := b_raw.strip()) else ""  # Only first line of description is relevant.
+
+        parts_a = a_raw.split()
         parts_b = b.split()
         section_key = parts_a[0].lower()
+
+        # Check for bad whitespaces in section definition
+        bad_whitespaces_def.extend(fetch_bad_ws_definitions(section, section_key, sep, a_raw, b_raw))
+
         if is_func and section_key in param_set:
             params.append(parse_section_param(section_key, sep, parts_a, order))
+
         elif is_func and section_key == ptype_key:
             params.append(parse_section_type(section_key, sep, parts_a, parts_b, order))
         elif is_func and section_key in rtype_key:
@@ -504,15 +562,25 @@ def parse_docs(node, /):
         elif section_key not in invalid:
             invalid.append(section_key)
 
+    non_empty_end_lines = False
+    if section_key in types_set and len(b_raw.splitlines()) > 1:
+        # Check for non-empty lines after the last section. (Only for type sections that cannot have a description)
+        non_empty_end_lines = any(line.strip() for line in b_raw.splitlines()[1:])  # Non-empty lines
+
     if docs:
         summary = get_summary(docs)
         first = node.body[0]
         docs_ini_lineno, docs_end_lineno = (first.lineno, first.end_lineno)
-        code_ini_lineno = node.body[1].lineno if len(node.body) > 1 else None
     else:
         summary = None
         docs_ini_lineno = docs_end_lineno = None
-        code_ini_lineno = node.body[0].lineno if node.body else None
+
+    code_ini_lineno = None
+    if docs_end_lineno:
+        after_lineno = docs_end_lineno + 1
+        after_line = linecache.getline(filename, after_lineno)
+        if after_line.strip() != "":  # Non-empty line after docstring
+            code_ini_lineno = after_lineno
 
     return ParsedDocs(
         kind=kind,
@@ -528,6 +596,8 @@ def parse_docs(node, /):
         variables=variables,
         invalid=invalid,
         ignored=ignored,
+        bad_whitespaces_def=bad_whitespaces_def,
+        non_empty_end_lines=non_empty_end_lines,
     )
 
 
@@ -618,18 +688,19 @@ def is_not_implemented(node, /, rawdocs=None):
             return False
 
 
-def checker(node, violations, /):
+def checker(node, violations, filename, /):
     """
     Explore the given AST node and yield each violation found.
 
     :param ast.AST node: Root node to explore.
     :param Violations violations: Violations instance with enabled/disabled rules.
+    :param str filename: Filename to check.
 
     :return: Generator with data of each violation.
     :rtype: typing.Iterator[tuple[int, str, dict]]
     """
 
-    parsed = parse_docs(node)
+    parsed = parse_docs(node, filename)
     lineno = parsed.docs_ini_lineno
 
     if parsed.kind == "function":
@@ -659,8 +730,8 @@ def get_params(node, /):
         yield ("return", ast.unparse(node.returns))
 
 
-def check_node(node, violations, /):
-    for lineno, code, msg, ctx in checker(node, violations):
+def check_node(node, violations, filename, /):
+    for lineno, code, msg, ctx in checker(node, violations, filename):
         yield (lineno, code, msg.format(*ctx))
 
 
@@ -717,7 +788,7 @@ def dump_statistics(violations, /):
 def dump_file(violations, quiet, path, /):
     def worker(content):
         for node in walk_module(quiet, content, filename):
-            yield from check_node(node, violations)
+            yield from check_node(node, violations, filename)
 
     getter = operator.itemgetter(0)  # lineno
     filename = str(path)
@@ -754,8 +825,8 @@ def main():
     parser.add_argument(
         "--ignore",
         nargs=argparse.ZERO_OR_MORE,
-        default=[".venv", ".env"],
-        help="directories to ignore",
+        default=default_ignore_dirs,
+        help="directories to ignore. Defaults to: {}".format(", ".join(default_ignore_dirs)),
     )
     parser.add_argument(
         "--enable",
