@@ -1,6 +1,7 @@
 # -*- coding: utf-8 -*-
 
 import ast
+import unittest.mock
 
 import pytest
 
@@ -16,6 +17,12 @@ def violations():
     return sphinxlinter.Violations()
 
 
+def chequer(content, violations, next_line=""):
+    with unittest.mock.patch.object(sphinxlinter.linecache, 'getline') as mock_getline:
+        mock_getline.return_value = next_line
+        yield from sphinxlinter.checker(parse_content(content), violations, 'dummy.py')
+
+
 def test_empty(violations):
     content = '''
 def foo():
@@ -23,7 +30,7 @@ def foo():
 
     pass
 '''
-    result = tuple(sphinxlinter.checker(parse_content(content), violations))
+    result = tuple(chequer(content, violations))
     assert not result
 
 
@@ -36,7 +43,7 @@ def foo():
 
     pass
 '''
-    result = tuple(sphinxlinter.checker(parse_content(content), violations))
+    result = tuple(chequer(content, violations))
     assert not result
 
 
@@ -64,7 +71,7 @@ def foo():
         (3, 'DOC001', 'Invalid docstring section ({!r})', ('cvar',)),
         (3, 'DOC001', 'Invalid docstring section ({!r})', ('vartype',)),
     )
-    result = tuple(sphinxlinter.checker(parse_content(content), violations))
+    result = tuple(chequer(content, violations))
     assert result == expected
 
 
@@ -83,7 +90,7 @@ def foo():
     expected = (
         (3, "DOC001", "Invalid docstring section ({!r})", ("foo",)),
     )
-    result = tuple(sphinxlinter.checker(parse_content(content), violations))
+    result = tuple(chequer(content, violations))
     assert result == expected
 
 
@@ -109,7 +116,7 @@ class Foo:
         (3, 'DOC001', 'Invalid docstring section ({!r})', ('rtype',)),
         (3, 'DOC001', 'Invalid docstring section ({!r})', ('var',)),
     )
-    result = tuple(sphinxlinter.checker(parse_content(content), violations))
+    result = tuple(chequer(content, violations))
     assert result == expected
 
 
@@ -125,7 +132,7 @@ def foo():
 
     pass
 '''
-    result = tuple(sphinxlinter.checker(parse_content(content), violations))
+    result = tuple(chequer(content, violations))
     assert not result
 
 
@@ -135,12 +142,10 @@ def foo():
     ("param", ":param int a",),  # missing separator ":" at the end when type is given
     ("param", ":param:",),  # missing parameter name without description and type
     ("param", ":param: description",),  # missing parameter name without type
-    ("param", ":param  : description",),  # missing parameter name (only spaces) without type
     # type
     ("type", ":type",),  # missing separator ":"
     ("type", ":type:",),  # missing separator name and type
     ("type", ":type:int",),  # missing parameter name
-    ("type", ":type :int",),  # missing parameter name (only spaces)
     ("type", ":type a:",),  # missing type
     ("type", ":type a: ",),  # missing type (only spaces)
     # return
@@ -172,7 +177,7 @@ def foo(a):
     expected = (
         (3, "DOC002", "Malformed section ({!r})", (section,)),
     )
-    result = tuple(sphinxlinter.checker(parse_content(content), violations))
+    result = tuple(chequer(content, violations))
     assert result == expected
 
 
@@ -187,8 +192,8 @@ def foo(a):
     ("ivar", ":ivar foo:",),
     ("cvar", ":cvar foo:",),
     # missing variable without name  and description (only spaces)
-    ("ivar", ":ivar  :  ",),
-    ("cvar", ":cvar  :  ",),
+    ("ivar", ":ivar:  ",),
+    ("cvar", ":cvar:  ",),
     # missing type
     ("vartype", ":vartype a:",),
     # missing type (only spaces)
@@ -208,25 +213,40 @@ class Foo:
     expected = (
         (3, "DOC002", "Malformed section ({!r})", (section,)),
     )
-    result = tuple(sphinxlinter.checker(parse_content(content), violations))
+    result = tuple(chequer(content, violations))
     assert result == expected
 
 
-def test_DOC003(violations):
+@pytest.mark.parametrize("next_line", ["pass", "# comment"])
+def test_DOC003(violations, next_line):
     content = '''
 def foo():
     """
     Title.
     """
-    pass
+    {next_line}
 '''
     expected = (
         (3, "DOC003", "Missing blank line after docstring", tuple()),
     )
-    result = tuple(sphinxlinter.checker(parse_content(content), violations))
+    result = tuple(chequer(content, violations, next_line=next_line))
     assert result == expected
 
 
+@pytest.mark.parametrize("next_line", ["", "\n"])
+def test_DOC003_ignore(violations, next_line):
+    content = '''
+def foo():
+    """
+    Title.
+    """
+'''
+    expected = ()
+    result = tuple(chequer(content, violations, next_line=next_line))
+    assert result == expected
+
+
+
 def test_DOC004(violations):
     content = '''
 def foo(a):
@@ -240,7 +260,7 @@ def foo(a):
     expected = (
         (3, "DOC004", "Missing blank line between summary and sections", ()),
     )
-    result = tuple(sphinxlinter.checker(parse_content(content), violations))
+    result = tuple(chequer(content, violations))
     assert result == expected
 
 
@@ -258,9 +278,9 @@ def foo(a):
 '''
 
     expected = (
-        (3, "DOC005", "Too many consecutive empty lines", ()),
+        (3, "DOC005", "Too many consecutive blank lines", ()),
     )
-    result = tuple(sphinxlinter.checker(parse_content(content), violations))
+    result = tuple(chequer(content, violations))
     assert result == expected
 
 
@@ -280,7 +300,7 @@ def foo(a):
     expected = (
         (3, "DOC006", "Trailing empty lines", ()),
     )
-    result = tuple(sphinxlinter.checker(parse_content(content), violations))
+    result = tuple(chequer(content, violations))
     assert result == expected
 
 
@@ -295,7 +315,7 @@ def foo():
     expected = (
         (3, "DOC006", "Trailing empty lines", ()),
     )
-    result = tuple(sphinxlinter.checker(parse_content(content), violations))
+    result = tuple(chequer(content, violations))
     assert result == expected
 
 
@@ -309,7 +329,7 @@ def foo(a):
     expected = (
         (3, "DOC008", "One-line docstring should end with a period", ()),
     )
-    result = tuple(sphinxlinter.checker(parse_content(content), violations))
+    result = tuple(chequer(content, violations))
     assert result == expected
 
 
@@ -327,7 +347,7 @@ def foo(a):
 '''
 
     expected = ()
-    result = tuple(sphinxlinter.checker(parse_content(content), violations))
+    result = tuple(chequer(content, violations))
     assert result == expected
 
 
@@ -345,7 +365,7 @@ def foo():
 
     expected = ((3, 'DOC009', 'Docstring must not use more than 3 double quotes', ()),)
 
-    result = tuple(sphinxlinter.checker(parse_content(content), violations))
+    result = tuple(chequer(content, violations))
     assert result == expected
 
 
@@ -361,7 +381,160 @@ def foo():
     pass
 '''
     expected = ()
-    result = tuple(sphinxlinter.checker(parse_content(content), violations))
+    result = tuple(chequer(content, violations))
+    assert result == expected
+
+
+@pytest.mark.parametrize("section, expected_section", [
+    # param type
+    (":type  a: str", ":type  a: str"),  # consecutive spaces between type and name
+    (": type a: str", ": type a: str"),  # leading space before type keyword
+    (":type a : str", ":type a : str"),  # trailing space after name
+    (":type a:  str", ":type a:  str"),  # leading space before type hint
+    (":type  a:  str", ":type  a:  str"),  # consecutive spaces both sides
+    # param
+    (":param  str  a: description", ":param  str  a:"),  # consecutive spaces
+    (": param str a: description", ": param str a:"),  # leading space before param keyword
+    (":param str a : description", ":param str a :"),  # trailing space after name
+    (":param  str a: description", ":param  str a:"),  # leading space before type hint
+    (":param str  a: description", ":param str  a:"),  # leading space before name
+    # return
+    (": return: description", ": return:"),  # leading space before return keyword
+    (":return : description", ":return :"),  # trailing space after return keyword
+    # # rtype
+    (": rtype: int", ": rtype: int"),  # leading space before rtype keyword
+    (":rtype : int", ":rtype : int"),  # trailing space after rtype keyword
+    (":rtype:  int", ":rtype:  int"),  # leading space before type hint
+    (": rtype :  int", ": rtype :  int"),  # leading and trailing space both sides
+    # # raises
+    (": raises ValueError: description", ": raises ValueError:"),  # leading ws before raises keyword
+    (":raises  ValueError: description", ":raises  ValueError:"),  # consecutive ws after raises keyword
+    (":raises ValueError : description", ":raises ValueError :"),  # trailing ws after error
+    (":raises ValueError,  KeyError: description", ":raises ValueError,  KeyError:"),  # Consecutive ws after comma
+    (":raises ValueError, KeyError : description", ":raises ValueError, KeyError :"),  # trailing ws after last error
+    (":raises ValueError  , KeyError: description", ":raises ValueError  , KeyError:"),  # leading ws before comma
+])
+def test_DOC010_function(section, violations, expected_section):
+    content = f'''
+def foo(a):
+    """
+    Title.
+
+    {section}
+    """
+
+    return a
+'''
+    expected = (
+        (3, "DOC010", "Section definition contains invalid whitespace ({!r})", (expected_section,)),
+    )
+    result = tuple(chequer(content, violations))
+    assert result == expected
+
+
+@pytest.mark.parametrize("section", [
+    ":param str a:  foo  bar  ",
+    ":return:  foo  bar  ",
+    ":raises ValueError:  foo  bar  ",
+])
+def test_DOC010_function_ignoring_descriptions_ws(section, violations):
+    content = f'''
+def foo(a):
+    """
+    Title.
+
+    {section}
+    """
+
+    return a
+'''
+    expected = ()
+    result = tuple(chequer(content, violations))
+    assert result == expected
+
+
+@pytest.mark.parametrize("section, expected_section", [
+    # var type
+    (":vartype  a: int", ":vartype  a: int"),  # consecutive spaces between vartype and name
+    (": vartype a: int", ": vartype a: int"),  # leading space before vartype keyword
+    (":vartype a : int", ":vartype a : int"),  # trailing space after name
+    (":vartype a:  int", ":vartype a:  int"),  # leading space before type hint
+    (":vartype  a:  int", ":vartype  a:  int"),  # consecutive spaces both sides
+    # ivar
+    (":ivar  a: description", ":ivar  a:"),  # consecutive spaces
+    (": ivar a: description", ": ivar a:"),  # leading space before ivar keyword
+    (":ivar a : description", ":ivar a :"),  # trailing space after name
+    # cvar
+    (":cvar  a: description", ":cvar  a:"),  # consecutive spaces
+    (": cvar a: description", ": cvar a:"),  # leading space before cvar keyword
+    (":cvar a : description", ":cvar a :"),  # trailing space after name
+])
+def test_DOC010_class(section, expected_section, violations):
+    content = f'''
+class Foo:
+    """
+    Title.
+
+    {section}
+    """
+
+    pass
+'''
+    expected = (
+        (3, "DOC010", "Section definition contains invalid whitespace ({!r})", (expected_section,)),
+    )
+    result = tuple(chequer(content, violations))
+    assert result == expected
+
+
+@pytest.mark.parametrize("section", [
+    ":ivar a:  foo  bar  ",
+    ":cvar a:  foo  bar  ",
+])
+def test_DOC010_class_ignoring_descriptions_ws(section, violations):
+    content = f'''
+class Foo:
+    """
+    Title.
+
+    {section}
+    """
+
+    pass
+'''
+    expected = ()
+    result = tuple(chequer(content, violations))
+    assert result == expected
+
+
+def test_DOC011_function(violations):
+    content = '''
+def foo():
+    """
+    Title.
+
+    :rtype: int
+    Trailing description.
+    """
+    '''
+
+    expected = ((3, 'DOC011', 'Trailing non-empty lines after last section.', ()),)
+    result = tuple(chequer(content, violations))
+    assert result == expected
+
+
+def test_DOC011_class(violations):
+    content = '''
+class Foo:
+    """
+    Title.
+
+    :vartype a: int
+    Trailing description.
+    """
+    '''
+    expected = ((3, 'DOC011', 'Trailing non-empty lines after last section.', ()),)
+    result = tuple(chequer(content, violations))
     assert result == expected
 
 
@@ -379,7 +552,7 @@ def foo():
     expected = (
         (3, "DOC101", "Parameter documented but not in signature ({!r})", ("a",)),
     )
-    result = tuple(sphinxlinter.checker(parse_content(content), violations))
+    result = tuple(chequer(content, violations))
     assert result == expected
 
 
@@ -397,7 +570,7 @@ def foo(a):
     expected = (
         ((3, "DOC102", "Invalid parameter type syntax ({!r})", ("list[str,",)),)
     )
-    result = tuple(sphinxlinter.checker(parse_content(content), violations))
+    result = tuple(chequer(content, violations))
     assert result == expected
 
 
@@ -415,7 +588,7 @@ def foo(a:int):
     expected = (
         ((3, "DOC103", "Parameter type already in signature ({!r})", ("int",)),)
     )
-    result = tuple(sphinxlinter.checker(parse_content(content), violations))
+    result = tuple(chequer(content, violations))
     assert result == expected
 
 
@@ -434,11 +607,11 @@ def foo(a:int):
         (
             3,
             "DOC104",
-            "Parameter type mismatch with hint ({!r} != {!r})",
+            "Parameter type mismatch with annotation ({!r} != {!r})",
             ("str", "int"),
         ),
     )
-    result = tuple(sphinxlinter.checker(parse_content(content), violations))
+    result = tuple(chequer(content, violations))
     assert result == expected
 
 
@@ -458,7 +631,7 @@ def foo(a):
     expected = (
         (3, "DOC105", "Duplicated parameter ({!r})", ("a",)),
     )
-    result = tuple(sphinxlinter.checker(parse_content(content), violations))
+    result = tuple(chequer(content, violations))
     assert result == expected
 
 
@@ -476,9 +649,9 @@ def foo(a, b):
     pass
 '''
     expected = (
-        (3, 'DOC007', 'Misplaced section ({!r} after {!r})', ('param', 'raises',)),
+        (3, 'DOC007', 'Misplaced section ({!r} appears after {!r})', ('param', 'raises',)),
     )
-    result = tuple(sphinxlinter.checker(parse_content(content), violations))
+    result = tuple(chequer(content, violations))
     assert result == expected
 
 
@@ -497,11 +670,11 @@ def foo(a, b, c):
     pass
 '''
     expected = (
-        (3, 'DOC007', 'Misplaced section ({!r} after {!r})', ('param', 'raises')),
-        (3, 'DOC007', 'Misplaced section ({!r} after {!r})', ('type', 'raises')),
-        (3, 'DOC007', 'Misplaced section ({!r} after {!r})', ('type', 'return')),
+        (3, 'DOC007', 'Misplaced section ({!r} appears after {!r})', ('param', 'raises')),
+        (3, 'DOC007', 'Misplaced section ({!r} appears after {!r})', ('type', 'raises')),
+        (3, 'DOC007', 'Misplaced section ({!r} appears after {!r})', ('type', 'return')),
     )
-    result = tuple(sphinxlinter.checker(parse_content(content), violations))
+    result = tuple(chequer(content, violations))
     assert result == expected
 
 
@@ -519,9 +692,9 @@ def foo():
 '''
 
     expected = (
-        (3, "DOC201", "Return documented but function has no return", ()),
+        (3, "DOC201", "Return documented but function has no return statement", ()),
     )
-    result = tuple(sphinxlinter.checker(parse_content(content), violations))
+    result = tuple(chequer(content, violations))
     assert result == expected
 
 
@@ -540,7 +713,7 @@ def foo():
     expected = (
         (3, "DOC202", "Invalid return type syntax ({!r})", ("list[int,",)),
     )
-    result = tuple(sphinxlinter.checker(parse_content(content), violations))
+    result = tuple(chequer(content, violations))
     assert result == expected
 
 
@@ -559,7 +732,7 @@ def foo() -> int:
     expected = (
         ((3, "DOC203", "Return type already in signature ({!r})", ("int",)),)
     )
-    result = tuple(sphinxlinter.checker(parse_content(content), violations))
+    result = tuple(chequer(content, violations))
     assert result == expected
 
 
@@ -578,7 +751,7 @@ def foo() -> int:
     expected = (
         (3, "DOC204", "Return type mismatch with annotation ({!r} != {!r})", ("str", "int")),
     )
-    result = tuple(sphinxlinter.checker(parse_content(content), violations))
+    result = tuple(chequer(content, violations))
     assert result == expected
 
 
@@ -599,7 +772,7 @@ def foo():
     expected = (
         ((3, "DOC205", "Duplicated return section ({!r})", (key,)),)
     )
-    result = tuple(sphinxlinter.checker(parse_content(content), violations))
+    result = tuple(chequer(content, violations))
     assert result == expected
 
 
@@ -618,7 +791,7 @@ def foo():
     expected = (
         ((3, "DOC302", "Invalid exception type syntax ({!r})", (True,)),)
     )
-    result = tuple(sphinxlinter.checker(parse_content(content), violations))
+    result = tuple(chequer(content, violations))
     assert result == expected
 
 
@@ -638,7 +811,7 @@ def foo():
     expected = (
         ((3, "DOC305", "Duplicated exception type ({!r})", ("ValueError",)),)
     )
-    result = tuple(sphinxlinter.checker(parse_content(content), violations))
+    result = tuple(chequer(content, violations))
     assert result == expected
 
 
@@ -657,9 +830,9 @@ def foo():
 '''
 
     expected = (
-        (3, "DOC007", "Misplaced section ({!r} after {!r})", ('raises', return_key)),
+        (3, "DOC007", "Misplaced section ({!r} appears after {!r})", ('raises', return_key)),
     )
-    result = tuple(sphinxlinter.checker(parse_content(content), violations))
+    result = tuple(chequer(content, violations))
     assert result == expected
 
 
@@ -678,7 +851,7 @@ class Foo:
     expected = (
         ((3, 'DOC402', 'Invalid variable type syntax ({!r})', ('list[str,',)),)
     )
-    result = tuple(sphinxlinter.checker(parse_content(content), violations))
+    result = tuple(chequer(content, violations))
     assert result == expected
 
 
@@ -698,8 +871,8 @@ class Foo:
 
     pass
 '''
-    expected = ((3, 'DOC403', 'Variable name must not contain spaces ({!r})', ('foo bar',)),)
-    result = tuple(sphinxlinter.checker(parse_content(content), violations))
+    expected = ((3, 'DOC403', 'Variable name contains invalid whitespace ({!r})', ('foo bar',)),)
+    result = tuple(chequer(content, violations))
     assert result == expected
 
 
@@ -716,5 +889,5 @@ class Foo:
     pass
 '''
     expected = ((3, 'DOC405', 'Duplicated variable ({!r})', ('bar',)),)
-    result = tuple(sphinxlinter.checker(parse_content(content), violations))
+    result = tuple(chequer(content, violations))
     assert result == expected
